package mediationpval;


import java.util.Arrays;
import java.util.concurrent.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import umontreal.iro.lecuyer.functionfit.SmoothingCubicSpline;
import umontreal.iro.lecuyer.gof.KernelDensity;
import umontreal.iro.lecuyer.probdist.EmpiricalDist;
import umontreal.iro.lecuyer.probdist.NormalDist;
import umontreal.iro.lecuyer.randvar.ChiSquareGen;
import umontreal.iro.lecuyer.randvar.NormalACRGen;
import umontreal.iro.lecuyer.rng.BasicRandomStreamFactory;
import umontreal.iro.lecuyer.rng.LFSR113;
import umontreal.iro.lecuyer.rng.MRG31k3p;
import umontreal.iro.lecuyer.rng.RandomStream;


/**
 *
 * @author falk
 *  
 *  Copyright 2012-2014 Carl F. Falk
 * 
 * This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
public class PartialPosteriorCalc {

    private double pValue; // main output variable of interest
    private double t1;
    private double t2;
    private int df1;
    private int df2;
    private int nNullsim;
    private int nGrid;
    private double qmin;
    private double qmax;
    private int rep;
    private int nDraws;
    private double rho;
    private RandomStream rstream;
    private NormalACRGen normalGen;
    private BasicRandomStreamFactory rstreamf;    
    private NormalDist nDist;
    private int progress;
    boolean done;

    private String calcMethod;
    private String approxMethod; // tdist or zdist


    // constructors
    public PartialPosteriorCalc(double t1, double t2, int df1, int df2,
            int nNullsim, int nGrid, double qmin, double qmax, int rep,
            int nDraws, double rho, String calcMethod, String approxMethod) {
        this.t1 = t1;
        this.t2 = t2;
        this.df1 = df1;
        this.df2 = df2;
        this.nNullsim = nNullsim;
        this.nGrid = nGrid;
        this.qmin = qmin;
        this.qmax = qmax;
        this.rep = rep;
        this.nDraws = nDraws;
        this.rho = rho;
        this.rstream = new MRG31k3p();
        this.normalGen = new NormalACRGen(new MRG31k3p(), new NormalDist());
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);        
        this.nDist = new NormalDist();
        this.progress = 0;
        this.done = false;
        this.calcMethod = calcMethod; // input not checked here; only 1 valid method
        this.approxMethod = approxMethod;
    }

    public PartialPosteriorCalc(double t1, double t2, int df1, int df2) {
        this.t1 = t1;
        this.t2 = t2;
        this.df1 = df1;
        this.df2 = df2;
        this.nNullsim = 1000000;
        this.nGrid = 250;
        this.qmin = -5;
        this.qmax = 5;
        this.rep = 1;
        this.nDraws = 50000;
        this.rho = .1;
        this.normalGen = new NormalACRGen(new MRG31k3p(), new NormalDist());
        this.rstream = new MRG31k3p();
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);        
        this.nDist = new NormalDist();
        this.progress = 0;
        this.done = false;

        this.calcMethod = "SmoothSpline";
        this.approxMethod = "tdist";
    }

    // getters/setters
    public int getDf1() {
        return df1;
    }

    public void setDf1(int df1) {
        this.df1 = df1;
    }

    public int getDf2() {
        return df2;
    }

    public void setDf2(int df2) {
        this.df2 = df2;
    }

    public double getT1() {
        return t1;
    }

    public void setT1(double t1) {
        this.t1 = t1;
    }

    public double getT2() {
        return t2;
    }

    public void setT2(double t2) {
        this.t2 = t2;
    }

    public int getnDraws() {
        return nDraws;
    }

    public void setnDraws(int nDraws) {
        this.nDraws = nDraws;
    }

    public int getnGrid() {
        return nGrid;
    }

    public void setnGrid(int nGrid) {
        this.nGrid = nGrid;
    }
    
    public double getqmin(){
        return qmin;
    }
    
    public void setqmin(double qmin){
        this.qmin = qmin;
    }
    
     public double getqmax(){
        return qmax;
    }
    
    public void setqmax(double qmax){
        this.qmax = qmax;
    }   

    public int getrep(){
        return rep;
    }
    
    public void setrep(int rep){
        this.rep = rep;
    }   

    
    public int getnNullsim() {
        return nNullsim;
    }

    public void setnNullsim(int nNullsim) {
        this.nNullsim = nNullsim;
    }

    private void setpValue(double pValue) {
        this.pValue = pValue;
    }

    public double getpValue() {
        return pValue;
    }

    public int getProgress(){
        return progress;
    }

    public void setProgress(int progress){
        this.progress = progress;
    }

    public String getCalcMethod(){
        return calcMethod;
    }

    public void setCalcMethod(String calcMethod){
        this.calcMethod = calcMethod;
    }

    public boolean getDone(){
        return done;
    }

    // calls default method for calculating p-value
    // implemented in this way in case a non monte-carlo method
    // or other approach implemented in future versions
    // for faster computation
    public void calcpVal() {
        this.done = false;
        double pvals = 0;
        setProgress(0);
        for(int i=0; i<rep; i++){
            if (calcMethod.equals("SmoothSpline")){
                pvals += this.getPPNull();
            }
        }
        this.setpValue(pvals/rep);
        this.done = true;
    }

    // functions that do the real work
    // only 2 calls here make this specific to work with t or z statistics:
    // getPosteriorT and getProdT
    //
    private double getPPNull() {

        double val; // return value of the function
        double[] grid = new double[nGrid]; // new grid
        double[] posterior = new double[nDraws]; // posterior of t1

        // get posterior dist for reg coefficient
        if(this.approxMethod.equals("tdist")){
            posterior = getPosteriorT(nDraws, t1, df1);
        } else if (this.approxMethod.equals("zdist")){
            posterior = getPosteriorZ(nDraws, t1);            
        }
        Arrays.sort(posterior);

        double[] xValues = new double[nGrid];
        double[] pValues = new double[nGrid];
        double[] dValues = new double[nGrid];
        double test = t1 * t2; // test statistic
        double[] prodT = new double[nNullsim]; // product of t distributions
        EmpiricalDist ecdf; // empirical distribution for product of t's
        double x; // points on the grid
        double x2[] = new double[1]; // point on the grid in vector form

        // this is not general; assumes some kind of smoothing over distribution
        // rather than re-calc at each posterior point
        for (int i = 0; i<nGrid; i++){
            grid[i] = t1+qmin+i*(qmax-qmin)/(nGrid-1);
        }
        
        for (int i = 0; i < nGrid; i++) {
            
            setProgress(getProgress()+1);
            x = grid[i];
            x2[0] = test;
            if(this.approxMethod.equals("tdist")){
                prodT = getProdT(nNullsim, x, 0, df1, df2);
            } else if (this.approxMethod.equals("zdist")){
                prodT = getProdZ(nNullsim, x, 0);                
            }
            Arrays.sort(prodT);
            ecdf = new EmpiricalDist(prodT);

            xValues[i] = x;
            pValues[i] = 1 - ecdf.cdf(Math.abs(test)) + ecdf.cdf(-1 * Math.abs(test));
            dValues[i] = KernelDensity.computeDensity(ecdf, nDist, x2)[0];

        }

            // type of smoother could be modified here
            val = splineEstimate (posterior, xValues, pValues, dValues);

        return val;
    }

    // estimation of p-value using splies
    // this is general enough that it could be used with z-score PP method
    private double splineEstimate (double[] posterior, double[] xValues,
            double[] pValues, double[] dValues) {
        double val;
        double dens; // density at a point
        double cdf; // p-value at a point
        double pDenom = 0;
        double pNum = 0;
        double minP;
        double minD;
        double[] pValCheck = new double[nGrid];
        double[] dValCheck = new double[nGrid];

        // get min p and d values
        System.arraycopy(pValues, 0, pValCheck, 0, nGrid);
        Arrays.sort(pValCheck);  
        minP = pValCheck[nGrid-1];
        System.arraycopy(dValues, 0, dValCheck, 0, nGrid);
        Arrays.sort(dValCheck);  
        minD = dValCheck[nGrid-1];     
        
        // alternative way to checking is to simply bound p-values and densities
        for(int i=0; i<nGrid; i++){
            if(pValues[i]<1e-50){ // 1e-50
                pValues[i]=1e-50;
            }
            if(dValues[i]<1e-50){
                dValues[i]=1e-50;
            }
        }
        
        if (minP > 1e-5 && minD!=0){
        
            SmoothingCubicSpline fit1 = new SmoothingCubicSpline(xValues, pValues, rho);
            SmoothingCubicSpline fit2 = new SmoothingCubicSpline(xValues, dValues, rho);

            for (int i = 0; i < nDraws; i++) {
                cdf = fit1.evaluate(posterior[i]);
                dens = fit2.evaluate(posterior[i]);
                
                // replacing 0 values with min observed value
                if(cdf == 0){
                    cdf = minP;
                }
                if(dens == 0){
                    dens = minD;
                }
                
                if (i == 0) {
                    pNum = (cdf / dens);
                    pDenom = 1 / dens;
                } else {
                    pNum += (cdf / dens);
                    pDenom += 1 / dens;
                }
            }

            val = pNum / pDenom; // this probably goes NaN when 0 in denom
            
            // ensure we don't get any invalid p-values
            if(val<0){
                val = 0.0;
            } else if (val>1){
                val = 1.0;
            }
        } else {
            val = 0.0;
        }
        return val;
    }

    private double getPosteriorT(double t, int df, RandomStream r) {
        double output;
        double z, cdf1, cdf11;
        z = this.normalGen.nextDouble();
        cdf1 = Math.sqrt(ChiSquareGen.nextDouble(r, df) / df);
        cdf11 = Math.sqrt(ChiSquareGen.nextDouble(r, df + 1) / df);        
        output = (z + (t) * (cdf1)) / (cdf11);
        return output;
    }

    private double[] getPosteriorT(int n, double t, int df) {
        double[] output = new double[n];
        
        // multicore processing
        int np = Runtime.getRuntime().availableProcessors(); 
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = n / np;
        int remainder = n % np;        
        double[] tmpp1;
        double[] tmpp2;
        double[] tmpp;        
        Callable m;
        Future f;
        
       // most of the result for p1
        for(int i=0; i<(np); i++){         
            m = new postTmulticore(section, t, df);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new postTmulticore(remainder,t,df);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, output, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, output, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
      
        multicore.shutdown();
        return output;
    }
    
    private double[] getPosteriorZ(int n, double z) {
        double[] output = new double[n];
        
        // multicore processing
        int np = Runtime.getRuntime().availableProcessors();        
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = n / np;
        int remainder = n % np;        
        double[] tmpp1;
        double[] tmpp2;
        double[] tmpp;        
        Callable m;
        Future f;
        
       // most of the result for p1
        for(int i=0; i<(np); i++){         
            m = new postZmulticore(section, z);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new postZmulticore(remainder,z);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, output, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, output, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
      
        multicore.shutdown();
        return output;
    }

    private double getProdT(double ncp1, double ncp2, int df1, int df2, RandomStream r) {
        double tStat1, tStat2;
        tStat1 = studentNoncentralGen(ncp1, df1, r);
        tStat2 = studentNoncentralGen(ncp2, df2, r);
        return tStat1 * tStat2;
    }

    private double[] getProdT(int n, double ncp1, double ncp2, int df1, int df2) {
        double[] output = new double[n];

        // multicore processing
        int np = Runtime.getRuntime().availableProcessors();        
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = n / np;
        int remainder = n % np;        
        double[] tmpp1;
        double[] tmpp2;
        double[] tmpp;        
        Callable m;
        Future f;
 
       // most of the result for p1
        for(int i=0; i<(np); i++){         
            m = new prodTmulticore(section, ncp1, ncp2, df1, df2);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new prodTmulticore(remainder,ncp1,ncp2,df1,df2);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, output, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, output, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
      
        multicore.shutdown();
        return output;
    }

    private double[] getProdZ(int n, double ncp1, double ncp2) {
        double[] output = new double[n];

        // multicore processing
        int np = Runtime.getRuntime().availableProcessors();        
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = n / np;
        int remainder = n % np;        
        double[] tmpp1;
        double[] tmpp2;
        double[] tmpp;        
        Callable m;
        Future f;
 
       // most of the result for p1
        for(int i=0; i<(np); i++){         
            m = new prodZmulticore(section, ncp1, ncp2);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new prodZmulticore(remainder,ncp1,ncp2);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, output, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, output, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(PartialPosteriorCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
      
        multicore.shutdown();
        return output;
    }
        
    private double studentNoncentralGen(double ncp, int df, RandomStream r) {
        double t, z, c;
        z = this.normalGen.nextDouble();
        c = Math.sqrt(ChiSquareGen.nextDouble(r, df) / df);
        
        t = (z + ncp) / c;
        return t;
    }
    
    class postTmulticore implements Callable {
        private int n;
        private double t;
        private int df;
        private double[] result;
        RandomStream r;
        
        public postTmulticore(int n, double t, int df){
            this.n=n;
            this.t=t;
            this.df=df;
            this.r= rstreamf.newInstance();               
        }
        
        @Override
        public double[] call(){
            result = new double[n];
            for(int i=0; i<n; i++){
                result[i] = getPosteriorT(t, df, r);
            }      
            return(result);
        }
    }    
    
    class postZmulticore implements Callable {
        private int n;
        private double z;
        private double[] result;
        NormalACRGen r;
        
        public postZmulticore(int n, double z){
            this.n=n;
            this.z=z;
            this.r= new NormalACRGen(rstreamf.newInstance(),z,1);               
        }
        
        @Override
        public double[] call(){
            result = new double[n];
            for(int i=0; i<n; i++){
                result[i] = r.nextDouble();
            }      
            return(result);
        }
    }
    
    class prodTmulticore implements Callable {
            private int n;
            private double ncp1;
            private double ncp2;
            private int df1;
            private int df2;
            private double[] result;
            RandomStream r;
            
              
        public prodTmulticore(int n, double ncp1, double ncp2, int df1, int df2){
            this.n=n;
            this.ncp1=ncp1;
            this.ncp2=ncp2;
            this.df1=df1;
            this.df2=df2;
            this.r = rstreamf.newInstance();            
        }
        
        @Override
        public double[] call(){
            
            result = new double[n];
            for(int i=0; i<n; i++){
                result[i] = getProdT(ncp1, ncp2, df1, df2, r);
            }      
            return(result);
        }
    }    
    
    class prodZmulticore implements Callable {
            private int n;
            private double ncp1;
            private double ncp2;
            private double[] result;
            NormalACRGen r1;
            NormalACRGen r2;            
            
              
        public prodZmulticore(int n, double ncp1, double ncp2){
            this.n=n;
            this.ncp1=ncp1;
            this.ncp2=ncp2;
            this.r1 = new NormalACRGen(rstreamf.newInstance(),ncp1,1);
            this.r2= new NormalACRGen(rstreamf.newInstance(),ncp2,1);               
            
        }
        
        @Override
        public double[] call(){
            
            result = new double[n];
            for(int i=0; i<n; i++){
                result[i] = r1.nextDouble()*r2.nextDouble();

            }      
            return(result);
        }
    }       
    
}
